import shlex
import shutil
import typing

from pydantic import Field, SecretStr

from .shell import Action as ShellAction

if typing.TYPE_CHECKING:
    pass


class Action(ShellAction):
    ssh_user: SecretStr
    ssh_host: SecretStr
    ssh_port: typing.Optional[int]
    ssh_password: SecretStr
    source: str
    destination: str
    options: list[str]
    excludes: list[str] = Field(default_factory=list)

    # _process: typing.Optional[pexpect.spawn] = PrivateAttr(None)

    def line_callback(self, line: str):
        if isinstance(line, str) and "Permission denied, please try again." in line:
            raise RuntimeError(f"{__name__}: Invalid password!")

    def build_args(self) -> list[str]:
        options = []
        for opt in self.options:
            list_opt = shlex.split(opt)
            options.extend(list_opt)

        rsh_opts = ""
        if self.ssh_port is not None and self.ssh_port != 22:
            rsh_opts += "-p " + str(self.ssh_port)

        if len(rsh_opts) > 1:
            options.append("--rsh")
            options.append("ssh " + rsh_opts + "")

        for exclude_pattern in self.excludes:
            options.append(f"--exclude")
            options.append(exclude_pattern)

        return options

    def initialize(self) -> None:
        super().initialize()
        command = shutil.which("rsync")
        if not command:
            raise RuntimeError(f"{__name__}: rsync command not found")

        args = self.build_args()
        args.append(self.source)
        destination = f"{self.ssh_user.get_secret_value()}@{self.ssh_host.get_secret_value()}:{self.destination}"
        args.append(destination)
        command = shlex.join([command, *args])
        with self.control_output():
            self.execute(command)
            self.execute(
                self.ssh_password.get_secret_value(),
                expect="password:",
                line_callback=self.line_callback,
            )



import functools
import typing

from click import Context as ClickContext
from rich import print, tree
from typer import Context, FileText, Option, Typer

from .schema import Step, Workflow

try:
    from dotenv import load_dotenv
except ImportError:
    load_dotenv = None

app = Typer(
    name="vonzy",
    help="Simple task runner for automation 😎",
    no_args_is_help=True,
)


def required_workflow(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        ctx: ClickContext = None
        for value in kwargs.values():
            if isinstance(value, ClickContext):
                ctx = value
                break
        if isinstance(ctx, ClickContext):
            workflow = ctx.obj
            if not isinstance(workflow, Workflow):
                print(
                    f"Error: no workflow found. use 'vonzy -c ...' option to load workflow file."
                )
                ctx.abort()
        return fn(*args, **kwargs)

    return wrapper


@app.callback(invoke_without_command=True)
def main(
    ctx: Context,
    config: FileText = Option(None, "-c", "--config", help="Configuration file"),
):
    if not config:
        return

    try:
        workflow = Workflow.parse_config(config)
        ctx.obj = workflow
    except Exception as e:
        print("Error:", e)
        ctx.abort()


@app.command()
@required_workflow
def run(
    ctx: Context,
    step_ids: typing.Optional[list[str]] = Option(
        None,
        "-s",
        "--step",
        help="Step IDs to run",
    ),
    env_file: typing.Optional[list[str]] = Option(
        None,
        "-e",
        "--env",
        help="Environment variables file (dotenv format)",
    ),
):
    """
    Run workflow
    """

    if load_dotenv is None and env_file:
        print(
            f"Error: you used the '-e' option but you haven't installed the python-dotenv module."
        )
        ctx.abort()

    for f in env_file:
        load_dotenv(f, override=True)

    workflow: Workflow = ctx.obj
    list(workflow.run(step_ids=step_ids))


@app.command()
@required_workflow
def steps(
    ctx: Context,
):
    """
    Show workflow steps
    """

    workflow: Workflow = ctx.obj
    comp = tree.Tree(f"List of steps in the {workflow.name!r} workflow")

    def _add_to_root(tr: tree.Tree, label: str, children: list[Step]):
        r = tr.add(label)
        for c in children:
            _add_to_root(r, c.name, c.steps)

    have_steps = len(workflow.steps) > 0
    for step in workflow.steps:
        if step.steps:
            _add_to_root(comp, step.name, step.steps)
        else:
            comp.add(step.name)

    if have_steps:
        print(comp)
    else:
        print(f"0 steps found in {workflow._source_file!r}")



import logging

from rich.logging import RichHandler

log = logging.getLogger("vonzy")
log.addHandler(RichHandler(rich_tracebacks=True, show_path=False))
log.setLevel(logging.NOTSET)
log.info("Starting vonzy")



import typing

import jinja2

from vonzy.logger import log

from .datatype import AttrDict

if typing.TYPE_CHECKING:
    from .schema import StepContext


def render_step_context(template: str, context: "StepContext") -> str:
    try:
        if template.startswith("{{") and template.endswith("}}"):
            rv = jinja2.Template(template).render(context.to_context())
        else:
            # use built-in str.format instead of string.Template.
            # By default python's str.format supports attribute fetching styles (aka, `getattr`) eg `obj.attr`.
            # While string.Template is not #cmiiw.
            # see: https://peps.python.org/pep-3101/#simple-and-compound-field-names
            kwargs = context.to_context()
            rv = template.format(**AttrDict(kwargs))
    except Exception as e:
        log.error(f"Error rendering template {template!r}: {e}")
        log.debug(f"Step context: {context}")
        raise

    return rv



import ast
import contextlib
import functools
import os
import re
import shutil
import typing

import pexpect
from pydantic import PrivateAttr

from ..logger import log
from ..utils import render_step_context
from .base import BaseAction

if typing.TYPE_CHECKING:
    from ..schema import StepContext

if os.name == "nt":
    raise RuntimeError("Windows is not supported.")

DEFAULT_SHELL = shutil.which("bash")


def clean(s):
    """
    Taken from: https://github.com/kennethreitz/crayons/blob/b1b78c9a357e0c348a1288ee5ef0318f08ccf257/crayons.py#L135C1-L139C15
    """
    strip = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")
    txt = strip.sub("", s)
    return txt  # .replace("\r", "").replace("\n", "")


def _validate_return_code(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        instance: Action = args[0]
        rv = fn(*args, **kwargs)
        if not instance.is_controlled:
            log.debug(f"Getting return code for cmd={args[1]}")
            results = fn(instance, cmd="echo $?", print_fn=None)
            log.debug(f"cmd={args[1]} results={results!r}")
            results = clean("".join(results))
            returncode = None
            try:
                returncode = ast.literal_eval(results)
            except Exception:
                pass

            if isinstance(returncode, int) and returncode != 0:
                raise RuntimeError(f"cmd={args[1]} returncode={returncode!r}")
        return rv

    return wrapper


class Action(BaseAction):
    cwd: typing.Optional[str] = None
    debug: typing.Optional[bool] = False
    timeout: typing.Optional[float] = 5
    _command: str = PrivateAttr(DEFAULT_SHELL)
    _process: typing.Optional[pexpect.pty_spawn.spawn] = PrivateAttr(None)
    _control_output: bool = PrivateAttr(False)
    _should_print: bool = PrivateAttr(False)

    def initialize(self) -> None:
        log.debug(f"Launches the command {self._command!r} into a background process.")
        if self._process is None:
            self._process = pexpect.spawn(
                self._command,
                encoding="utf-8",
                timeout=self.timeout,
                cwd=self.cwd,
                env=os.environ,
            )
            self._process.delaybeforesend = 0.1
            self._process.delayafterread = 0.1
            # self._process.logfile = sys.stdout

    def cleanup(self) -> typing.Optional[int]:
        log.debug(f"Closing process {self._command!r}")
        exit_code = 0
        if self._process:
            self._process.close()
            exit_code = self._process.exitstatus
            self._process = None
        log.debug(f"Process {self._command!r} exited with code {exit_code!r}")
        return exit_code

    @contextlib.contextmanager
    def control_output(self):
        self._control_output = True
        try:
            yield self
        finally:
            self._control_output = False

    @property
    def is_controlled(self) -> bool:
        return self._control_output

    @_validate_return_code
    def execute(
        self,
        cmd: str,
        *,
        expect: typing.Optional[str] = None,
        context: typing.Optional["StepContext"] = None,
        line_callback: typing.Optional[typing.Callable[[str], None]] = None,
        print_fn: typing.Optional[typing.Callable] = print,
    ):
        if not isinstance(cmd, str):
            raise RuntimeError(f"{__name__}: Command {cmd!r} is not a string.")

        if context is not None:
            cmd = render_step_context(cmd.strip(), context=context)

        default_expect = ["\r\n", "\n"]
        if expect:
            log.debug(f"Expects {expect!r} on stdin")
            self._process.expect(expect)
        else:
            # Hide the log if it has the `expect` param. To prevent displaying unwanted text/data on the console.
            log.debug(f"Executing command {cmd!r}")

        self._process.sendline(cmd)
        lines = []
        _should_print = self._should_print if self.is_controlled else False
        while True:
            try:
                self._process.expect(default_expect)
                line = self._process.before + self._process.after
                # print(f"should_print={_should_print} match_idx={match_idx!r} line={line!r}")
                if _should_print:
                    if isinstance(line, str):
                        line = clean(line)
                    if self.debug and callable(print_fn):
                        print_fn(line, end="")
                    if callable(line_callback):
                        line_callback(line)
                    lines.append(line)

                if self._process.before and self._process.before.startswith(
                    "\x1b[?2004h"
                ):
                    _should_print = True
                    if self.is_controlled:
                        self._should_print = _should_print
            except (pexpect.TIMEOUT, pexpect.EOF) as e:
                # log.error(f"Command {cmd!r} timed out/eof. {e}")
                break

        return lines



import typing

from pydantic import PrivateAttr, SecretStr

from .base import BaseAction

if typing.TYPE_CHECKING:
    from ..schema import StepContext

from ..utils import render_step_context

try:
    import paramiko
except ImportError:
    raise ImportError("paramiko module not found. try: pip install paramiko")


class Action(BaseAction):
    ssh_host: SecretStr
    ssh_user: SecretStr
    ssh_port: typing.Optional[int] = 22
    ssh_password: SecretStr

    _ssh_client: typing.Optional[paramiko.SSHClient] = PrivateAttr(None)

    def handle_commands(
        self,
        commands: list[typing.Any],
        *,
        context: typing.Optional["StepContext"] = None,
    ) -> None:
        cmd_list = []
        for cmd in commands:
            if not isinstance(cmd, str):
                raise RuntimeError(f"{__name__}: Command {cmd!r} is not a string.")

            cmd = render_step_context(cmd.strip(), context=context)
            cmd_list.append(cmd)

        return [";".join(cmd_list)]

    def initialize(self) -> None:
        self._ssh_client = paramiko.SSHClient()
        self._ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            self._ssh_client.connect(
                hostname=self.ssh_host.get_secret_value(),
                port=self.ssh_port,
                username=self.ssh_user.get_secret_value(),
                password=self.ssh_password.get_secret_value(),
            )
        except Exception as e:
            raise RuntimeError(f"{__name__}: {e}")

    def cleanup(self):
        try:
            self._ssh_client.close()
        except AttributeError as e:
            if "'NoneType' object has no attribute 'time'" in str(e):
                pass

        self._ssh_client = None

    def execute(
        self,
        cmd: str,
        *,
        context: typing.Optional["StepContext"] = None,
    ) -> None:
        stdin, stdout, stderr = self._ssh_client.exec_command(cmd)
        stdin.close()
        stdout.channel.set_combine_stderr(True)
        for line in stdout:
            print(line, end="")

        returncode = stdout.channel.recv_exit_status()
        if isinstance(returncode, int) and returncode != 0:
            raise RuntimeError(f"{__name__}: cmd={cmd!r} returncode={returncode!r}")



import logging
import os
import string
from ast import literal_eval
from enum import Enum
from importlib import import_module
from io import BufferedReader
from typing import Any, Callable, Literal, Optional, Type, TypeVar, Union

import oyaml as yaml
from inquirer import Checkbox, List, Password, Text, prompt
from inquirer.questions import Question
from jinja2 import Template
from pydantic import BaseModel, Field, PrivateAttr, validator

from . import actions
from .actions.base import BaseAction
from .constants import BASE_RULE_JINJA_TEMPLATE
from .datatype import AttrDict
from .errors import InvalidAction, InvalidStep, MissingDependency
from .logger import log
from .utils import render_step_context

try:
    from dotenv import load_dotenv
except ImportError:
    load_dotenv = None  # type: ignore[assignment]

try:
    from pydantic import EmailStr
except ImportError:
    EmailStr = None  # type: ignore[assignment, misc]

T = TypeVar("T")


class Input(BaseModel):
    class Types(str, Enum):
        text = "text"
        email = "email"
        password = "password"
        number = "number"
        list = "list"
        checkbox = "checkbox"

    key: str
    value: Optional[str]
    type: Types
    required: Optional[bool]
    default: Optional[str]
    choices: list[str] = Field(default_factory=list)
    description: str

    def get_value_validator(self) -> Optional[Callable[[str], Any]]:
        fn: Optional[Callable[[str], Any]] = str
        if self.type in (self.Types.checkbox, self.Types.list):
            fn = None  # type: ignore[assignment]
        elif self.type == self.Types.email:
            if EmailStr is None:
                raise MissingDependency("email-validator")
            fn = EmailStr
        elif self.type == self.Types.number:
            fn = int

        def validator(_, current: Any) -> bool:
            try:
                fn(current)
            except Exception:
                return False
            return True

        if fn:
            return validator

    def get_real_value(self) -> Optional[str]:
        v = self.value
        if v is None:
            v = self.default
        return v

    def get_widget(self) -> tuple[Type[Question], dict]:
        widget_obj = Text
        widget_params = {
            "name": self.key,
            "message": self.description,
            "default": self.get_real_value(),
            "ignore": lambda x: False if self.required else None,
            "validate": self.get_value_validator(),
        }
        valid_type = False
        if self.type in (self.Types.text, self.Types.email, self.Types.number):
            valid_type = True
        elif self.type == self.Types.password:
            valid_type = True
            widget_obj = Password
        elif self.type in (self.Types.list, self.Types.checkbox):
            valid_type = True
            widget_obj = Checkbox if self.type == self.Types.checkbox else List
            widget_params.pop("validate")
            widget_params["choices"] = self.choices

        if not valid_type:
            raise InvalidAction(f"Unknown input type {self.type!r}")

        return widget_obj, widget_params


class Action(BaseModel):
    name: str
    params: Optional[dict] = Field(default_factory=dict)
    klass: Optional[str] = Field("Action")
    _instance: Optional[BaseAction] = PrivateAttr(None)


class CommandRule(BaseModel):
    rule: str
    cmd: Union[str, dict]


class Step(BaseModel):
    id: str
    name: str
    use: Union[Action, str]
    rule: Optional[str]
    commands: list[Union[str, CommandRule]] = Field(default_factory=list)
    steps: list["Step"] = Field(default_factory=list)

    @validator("id", always=True)
    def validate_id(cls, v: str):
        if v == "result":
            raise InvalidStep(
                f"The step cannot have id='result' as this will be used to store the result of the step."
            )
        return v

    def load_action(self, sc: "StepContext") -> Action:
        use_action = self.use
        if isinstance(use_action, str):
            use_action = Action(name=use_action)

        log.info(f"Loading action {use_action.name!r}")
        try:
            action_name = use_action.name.lstrip(".")
            action_class = actions.__cached_actions__.get(action_name)
            if not action_class:
                action_module = import_module(action_name)
                action_class: Optional[BaseAction] = getattr(
                    action_module, use_action.klass, None
                )
                if action_class is None:
                    raise InvalidAction(f"Action object not found in {action_name!r}")
                else:
                    actions.__cached_actions__[action_name] = action_class

            log.info(f"Action {action_name!r} loaded")
            action_params = {}
            for k, v in (use_action.params or {}).items():
                if isinstance(v, str):
                    v = render_step_context(v, context=sc)
                if isinstance(v, list):
                    nv = []
                    for vv in v:
                        if isinstance(vv, str):
                            vv = render_step_context(vv, context=sc)
                        nv.append(vv)
                    v = nv
                action_params[k] = v

            action_instance = action_class(**action_params)
            log.debug(f"Action {action_instance} initialized")
            use_action._instance = action_instance
            return use_action

        except ImportError:
            raise InvalidAction(f"Action {action_name!r} not found")

    def _validate_rule(self, expr: str, sc: "StepContext") -> bool:
        rule_expr = string.Template(BASE_RULE_JINJA_TEMPLATE).substitute(expr=expr)
        result = Template(rule_expr).render(sc.to_context())
        rv = literal_eval(result)
        return rv

    def run(self, sc: "StepContext", *, parent_step_ids: Optional[list[str]] = None):
        step_id = self.id
        steps_ctx = sc.steps
        if not isinstance(steps_ctx, AttrDict):
            steps_ctx = AttrDict()
            sc.steps = steps_ctx

        sc.steps = steps_ctx
        _current_step_data: Optional[AttrDict] = None
        if parent_step_ids:
            for sid in parent_step_ids:
                if _current_step_data:
                    _current_step_data = _current_step_data.get(sid)
                else:
                    _current_step_data = steps_ctx.get(sid)

                if _current_step_data is None:
                    raise InvalidStep(f"Step {sid!r} not found -> {parent_step_ids}")

        if not _current_step_data:
            _current_step_data = steps_ctx

        realname = render_step_context(self.name, context=sc)
        self.name = realname
        log.info(f"Running step {self.name!r} #{step_id}")
        result_class = StepResult
        if self.rule:
            rule_passed = self._validate_rule(self.rule, sc)
            if not rule_passed:
                log.info(f"Step {self.id!r} skipped")
                result = result_class(step=self, status="skipped", value=None)
                _current_step_data[step_id] = AttrDict(result=result)
                yield result
                return

        action_obj = self.load_action(sc)
        result = None
        try:
            action_obj._instance.initialize()
            commands = action_obj._instance.handle_commands(self.commands, context=sc)
            for cmd in commands:
                if isinstance(cmd, CommandRule):
                    if not self._validate_rule(cmd.rule, sc):
                        log.debug(f"Command {cmd.cmd!r} skipped.")
                        continue
                    cmd = cmd.cmd

                kwargs = {}
                args = (cmd,)
                kwargs["context"] = sc
                action_obj._instance.execute(*args, **kwargs)
            result = result_class(step=self, status="success", value=None)
        except Exception as e:
            result = result_class(step=self, status="error", value=e)
        finally:
            try:
                action_obj._instance.cleanup()
            except Exception as e:
                log.error(
                    f"Error cleaning up action {action_obj.name!r} on step {self.id!r}: {e}"
                )
                result = result_class(step=self, status="error", value=e)

        _R = result.dict(exclude={"step"})
        log.info(f"Step {self.id!r} finished with status={_R['status']}")
        log.debug(f"Result {_R} for step {self.id!r}")
        _current_step_data[step_id] = AttrDict(result=result)
        yield result
        for child_step in self.steps:
            if parent_step_ids is None:
                parent_step_ids = [self.id]
            else:
                parent_step_ids.append(self.id)

            for child_result in child_step.run(sc, parent_step_ids=parent_step_ids):
                yield child_result


class StepResult(BaseModel):
    step: Step
    status: Literal["success", "error", "skipped"]
    value: Optional[Any]


class StepContext(BaseModel):
    env: AttrDict
    inputs: Optional[AttrDict]
    steps: Optional[AttrDict[str, StepResult]]

    def to_context(self) -> dict[str, Any]:
        return {
            "env": self.env,
            "inputs": self.inputs,
            "steps": self.steps,
        }


class Workflow(BaseModel):
    name: str
    log_level: str = "NOTSET"
    env_file: Optional[Union[str, list[str]]] = None
    inputs: list[Input] = Field(default_factory=list)
    steps: list[Step] = Field(default_factory=list)

    _source_file: Optional[str] = PrivateAttr(None)

    @validator("log_level")
    def _validate_and_set_log_level(cls, v: str):
        v = v.upper()
        if v not in logging._nameToLevel:
            raise ValueError(f"Invalid log level {v!r}")
        log.setLevel(v)
        return v

    def before_run(self, sc: StepContext):
        questions: list[Question] = []
        for input in self.inputs:
            widget_class, widget_params = input.get_widget()
            default = widget_params["default"]
            if isinstance(default, str):
                default = render_step_context(default, context=sc)

            widget_params["default"] = default
            widget = widget_class(**widget_params)
            questions.append(widget)

        values = prompt(questions, raise_keyboard_interrupt=True)
        return values

    def load_env_file(self):
        if load_dotenv is not None and self.env_file:
            env_file = self.env_file
            if not isinstance(env_file, list):
                env_file = [env_file]

            for f in env_file:
                load_dotenv(f)

    @classmethod
    def parse_config(cls, config: BufferedReader):
        data = yaml.safe_load(config)
        instance = cls(**data)
        instance._source_file = config.name
        return instance

    @classmethod
    def load_config(cls, src: str):
        with open(src, "rb") as f:
            return cls.parse_config(f)

    def run(self, step_ids: Optional[list[str]] = None):
        self.load_env_file()
        try:
            ctx = StepContext(
                env=AttrDict(**os.environ),
            )
            inputs_ctx = self.before_run(ctx)
            if inputs_ctx:
                ctx.inputs = inputs_ctx

            ctx.steps = AttrDict()
            have_steps_ids = isinstance(step_ids, list)
            for step in self.steps:
                if have_steps_ids and step.id not in step_ids:
                    ctx.steps[step.id] = AttrDict(
                        result=StepResult(step=step, status="skipped", value=None)
                    )
                    continue
                for result in step.run(ctx):
                    yield result
        except KeyboardInterrupt:
            log.info("Cancelled by user.")
        except Exception as e:
            log.error(f"Error: {e}", exc_info=True)

        yield



import typing
from abc import ABC, abstractmethod

from pydantic import BaseModel

if typing.TYPE_CHECKING:
    from vonzy.schema import StepContext

T = typing.TypeVar("T")


class BaseAction(ABC, BaseModel):
    @abstractmethod
    def initialize(self) -> None:
        pass

    @abstractmethod
    def cleanup(self):
        pass

    @abstractmethod
    def execute(
        self, *args, context: typing.Optional[dict[typing.Any, typing.Any]] = None
    ) -> None:
        pass

    def handle_commands(
        self, commands: T, *, context: typing.Optional["StepContext"] = None
    ) -> T:
        return commands
